/* $$ Clover has instrumented this file $$ */// Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tapestry.Tapestry;

/**
 *  An implementation of the <b>Adaptor</b> pattern.  The adaptor
 *  pattern allows new functionality to be assigned to an existing class.
 *  As implemented here, this is a smart lookup between
 *  a particular class (the class to be adapted, called
 *  the <em>subject class</em>) and some object instance
 *  that can provide the extra functionality (called the
 *  <em>adaptor</em>).  The implementation of the adaptor is not relevant
 *  to the AdaptorRegistry class.
 *
 *  <p>Adaptors are registered before they can be used; the registration maps a
 *  particular class to an adaptor instance.  The adaptor instance will be used
 *  when the subject class matches the registered class, or the subject class
 *  inherits from the registered class.
 *
 *  <p>This means that a search must be made that walks the inheritance tree
 *  (upwards from the subject class) to find a registered mapping.
 *
 *  <p>In addition, adaptors can be registered against <em>interfaces</em>.
 *  Searching of interfaces occurs after searching of classes.  The exact order is:
 *
 *  <ul>
 *  <li>Search for the subject class, then each super-class of the subject class
 *      (excluding java.lang.Object)
 *  <li>Search interfaces, starting with interfaces implemented by the subject class,
 *  continuing with interfaces implemented by the super-classes, then
 *  interfaces extended by earlier interfaces (the exact order is a bit fuzzy)
 *  <li>Search for a match for java.lang.Object, if any
 *  </ul>
 *
 *  <p>The first match terminates the search.
 *
 *  <p>The AdaptorRegistry caches the results of search; a subsequent search for the
 *  same subject class will be resolved immediately.
 * 
 *  <p>AdaptorRegistry does a minor tweak of the "natural" inheritance.
 *  Normally, the parent class of an object array (i.e., <code>Foo[]</code>) is
 *  simply <code>Object</code>, even though you may assign 
 *  <code>Foo[]</code> to a variable of type <code>Object[]</code>.  AdaptorRegistry
 *  "fixes" this by searching for <code>Object[]</code> as if it was the superclass of
 *  any object array.  This means that the search path for <code>Foo[]</code> is
 *  <code>Foo[]</code>, <code>Object[]</code>, then a couple of interfaces 
 *  {@link java.lang.Cloneable}, {@link java.io.Serializable}, etc. that are\
 *  implicitily implemented by arrarys), and then, finally, <code>Object</code>
 * 
 *  <p>
 *  This tweak doesn't apply to scalar arrays, since scalar arrays may <em>not</em>
 *  be assigned to <code>Object[]</code>. 
 *
 *  <p>This class is thread safe.
 *
 *
 *  @author Howard Lewis Ship
 * 
 **/

public class AdaptorRegistry
{public static com.cortexeb.tools.clover.d __CLOVER_389_0 = com.cortexeb.tools.clover.aq.getRecorder(new char[] {67,58,92,119,111,114,107,115,112,97,99,101,92,106,97,107,97,114,116,97,45,116,97,112,101,115,116,114,121,92,102,114,97,109,101,119,111,114,107,92,116,97,114,103,101,116,92,99,108,111,118,101,114,45,100,98},1096998272901L);
    private static final Log LOG = LogFactory.getLog(AdaptorRegistry.class);

    /**
     *  A Map of adaptor objects, keyed on registration Class.
     *
     **/

    private Map registrations = new HashMap();

    /**
     *  A Map of adaptor objects, keyed on subject Class.
     *
     **/

    private Map cache = new HashMap();

    /**
     *  Registers an adaptor for a registration class.
     *
     *  @throws IllegalArgumentException if an adaptor has already
     *  been registered for the given class.
     **/

    public synchronized void register(Class registrationClass, Object adaptor)
    {try { __CLOVER_389_0.M[2112]++;
        __CLOVER_389_0.S[8659]++;if ((((registrations.containsKey(registrationClass)) && (++__CLOVER_389_0.CT[1448] != 0)) || (++__CLOVER_389_0.CF[1448] == 0))){
            __CLOVER_389_0.S[8660]++;throw new IllegalArgumentException(
                Tapestry.format(
                    "AdaptorRegistry.duplicate-registration",
                    Tapestry.getClassName(registrationClass)));}

        __CLOVER_389_0.S[8661]++;registrations.put(registrationClass, adaptor);

        __CLOVER_389_0.S[8662]++;if ((((LOG.isDebugEnabled()) && (++__CLOVER_389_0.CT[1449] != 0)) || (++__CLOVER_389_0.CF[1449] == 0))){
            __CLOVER_389_0.S[8663]++;LOG.debug("Registered " + adaptor + " for " + Tapestry.getClassName(registrationClass));}

        // Can't tell what is and isn't valid in the cache.
        // Also, normally all registrations occur before any adaptors
        // are searched for, so this is not a big deal.

        __CLOVER_389_0.S[8664]++;cache.clear();
    } finally { }}

    /**
     *  Gets the adaptor for the specified subjectClass.
     *
     *  @throws IllegalArgumentException if no adaptor could be found.
     *
     **/

    public synchronized Object getAdaptor(Class subjectClass)
    {try { __CLOVER_389_0.M[2113]++;
        __CLOVER_389_0.S[8665]++;Object result;

        __CLOVER_389_0.S[8666]++;if ((((LOG.isDebugEnabled()) && (++__CLOVER_389_0.CT[1450] != 0)) || (++__CLOVER_389_0.CF[1450] == 0))){
            __CLOVER_389_0.S[8667]++;LOG.debug("Getting adaptor for class " + Tapestry.getClassName(subjectClass));}

        __CLOVER_389_0.S[8668]++;result = cache.get(subjectClass);

        __CLOVER_389_0.S[8669]++;if ((((result != null) && (++__CLOVER_389_0.CT[1451] != 0)) || (++__CLOVER_389_0.CF[1451] == 0))){
        {
            __CLOVER_389_0.S[8670]++;if ((((LOG.isDebugEnabled()) && (++__CLOVER_389_0.CT[1452] != 0)) || (++__CLOVER_389_0.CF[1452] == 0))){
                __CLOVER_389_0.S[8671]++;LOG.debug("Found " + result + " in cache");}

            __CLOVER_389_0.S[8672]++;return result;
        }}

        __CLOVER_389_0.S[8673]++;result = searchForAdaptor(subjectClass);

        // Record the result in the cache

        __CLOVER_389_0.S[8674]++;cache.put(subjectClass, result);

        __CLOVER_389_0.S[8675]++;if ((((LOG.isDebugEnabled()) && (++__CLOVER_389_0.CT[1453] != 0)) || (++__CLOVER_389_0.CF[1453] == 0))){
            __CLOVER_389_0.S[8676]++;LOG.debug("Found " + result);}

        __CLOVER_389_0.S[8677]++;return result;
    } finally { }}

    /**
     * Searches the registration Map for a match, based on inheritance.
     *
     * <p>Searches class inheritance first, then interfaces (in a rather vague order).
     * Really should match the order from the JVM spec.
     *
     * <p>There's a degenerate case where we may check the same interface more than once:
     * <ul>
     * <li>Two interfaces, I1 and I2
     * <li>Two classes, C1 and C2
     * <li>I2 extends I1
     * <li>C2 extends C1
     * <li>C1 implements I1
     * <li>C2 implements I2
     * <li>The search will be: C2, C1, I2, I1, I1
     * <li>I1 is searched twice, because C1 implements it, and I2 extends it
     * <li>There are other such cases, but none of them cause infinite loops
     * and most are rare (we could guard against it, but its relatively expensive).
     * <li>Multiple checks only occur if we don't find a registration
     * </ul>
     *
     *  <p>
     *  This method is only called from a synchronized block, so it is
     *  implicitly synchronized.
     * 
     **/

    private Object searchForAdaptor(Class subjectClass)
    {try { __CLOVER_389_0.M[2114]++;
        __CLOVER_389_0.S[8678]++;LinkedList queue = null;
        __CLOVER_389_0.S[8679]++;Object result = null;

        __CLOVER_389_0.S[8680]++;if ((((LOG.isDebugEnabled()) && (++__CLOVER_389_0.CT[1454] != 0)) || (++__CLOVER_389_0.CF[1454] == 0))){
            __CLOVER_389_0.S[8681]++;LOG.debug("Searching for adaptor for class " + Tapestry.getClassName(subjectClass));}

        // Step one: work up through the class inheritance.

        __CLOVER_389_0.S[8682]++;Class searchClass = subjectClass;

        // Primitive types have null, not Object, as their parent
        // class.

        __CLOVER_389_0.S[8683]++;while ((((searchClass != Object.class && searchClass != null) && (++__CLOVER_389_0.CT[1455] != 0)) || (++__CLOVER_389_0.CF[1455] == 0))){
        {
            __CLOVER_389_0.S[8684]++;result = registrations.get(searchClass);
            __CLOVER_389_0.S[8685]++;if ((((result != null) && (++__CLOVER_389_0.CT[1456] != 0)) || (++__CLOVER_389_0.CF[1456] == 0))){
                __CLOVER_389_0.S[8686]++;return result;}

            // Not an exact match.  If the search class
            // implements any interfaces, add them to the queue.

            __CLOVER_389_0.S[8687]++;Class[] interfaces = searchClass.getInterfaces();
            __CLOVER_389_0.S[8688]++;int length = interfaces.length;

            __CLOVER_389_0.S[8689]++;if ((((queue == null && length > 0) && (++__CLOVER_389_0.CT[1457] != 0)) || (++__CLOVER_389_0.CF[1457] == 0))){
                __CLOVER_389_0.S[8690]++;queue = new LinkedList();}

            __CLOVER_389_0.S[8691]++;for (int i = 0; (((i < length) && (++__CLOVER_389_0.CT[1458] != 0)) || (++__CLOVER_389_0.CF[1458] == 0)); i++){
                __CLOVER_389_0.S[8692]++;queue.addLast(interfaces[i]);}

            // Advance up to the next superclass

            __CLOVER_389_0.S[8693]++;searchClass = getSuperclass(searchClass);

        }}

        // Ok, the easy part failed, lets start searching
        // interfaces.

        __CLOVER_389_0.S[8694]++;if ((((queue != null) && (++__CLOVER_389_0.CT[1459] != 0)) || (++__CLOVER_389_0.CF[1459] == 0))){
        {
            __CLOVER_389_0.S[8695]++;while ((((!queue.isEmpty()) && (++__CLOVER_389_0.CT[1460] != 0)) || (++__CLOVER_389_0.CF[1460] == 0))){
            {
                __CLOVER_389_0.S[8696]++;searchClass = (Class) queue.removeFirst();

                __CLOVER_389_0.S[8697]++;result = registrations.get(searchClass);
                __CLOVER_389_0.S[8698]++;if ((((result != null) && (++__CLOVER_389_0.CT[1461] != 0)) || (++__CLOVER_389_0.CF[1461] == 0))){
                    __CLOVER_389_0.S[8699]++;return result;}

                // Interfaces can extend other interfaces; add them
                // to the queue.

                __CLOVER_389_0.S[8700]++;Class[] interfaces = searchClass.getInterfaces();
                __CLOVER_389_0.S[8701]++;int length = interfaces.length;

                __CLOVER_389_0.S[8702]++;for (int i = 0; (((i < length) && (++__CLOVER_389_0.CT[1462] != 0)) || (++__CLOVER_389_0.CF[1462] == 0)); i++){
                    __CLOVER_389_0.S[8703]++;queue.addLast(interfaces[i]);}
            }}
        }}

        // Not a match on interface; our last gasp is to check
        // for a registration for java.lang.Object

        __CLOVER_389_0.S[8704]++;result = registrations.get(Object.class);
        __CLOVER_389_0.S[8705]++;if ((((result != null) && (++__CLOVER_389_0.CT[1463] != 0)) || (++__CLOVER_389_0.CF[1463] == 0))){
            __CLOVER_389_0.S[8706]++;return result;}

        // No match?  That's rare ... and an error.

        __CLOVER_389_0.S[8707]++;throw new IllegalArgumentException(
            Tapestry.format(
                "AdaptorRegistry.adaptor-not-found",
                Tapestry.getClassName(subjectClass)));
    } finally { }}

    /**
     *  Returns the superclass of the given class, with a single tweak:  If the 
     *  search class is an array class, and the component type is an object class
     *  (but not Object), then the simple Object array class is returned.  This reflects
     *  the fact that an array of any class may be assignable to <code>Object[]</code>,
     *  even though the superclass of an array is always simply <code>Object</code>.
     * 
     **/

    private Class getSuperclass(Class searchClass)
    {try { __CLOVER_389_0.M[2115]++;
        __CLOVER_389_0.S[8708]++;if ((((searchClass.isArray()) && (++__CLOVER_389_0.CT[1464] != 0)) || (++__CLOVER_389_0.CF[1464] == 0))){
        {
            __CLOVER_389_0.S[8709]++;Class componentType = searchClass.getComponentType();

            __CLOVER_389_0.S[8710]++;if ((((!componentType.isPrimitive() && componentType != Object.class) && (++__CLOVER_389_0.CT[1465] != 0)) || (++__CLOVER_389_0.CF[1465] == 0))){
                __CLOVER_389_0.S[8711]++;return Object[].class;}
        }}

        __CLOVER_389_0.S[8712]++;return searchClass.getSuperclass();
    } finally { }}

    public synchronized String toString()
    {try { __CLOVER_389_0.M[2116]++;
        __CLOVER_389_0.S[8713]++;StringBuffer buffer = new StringBuffer();
        __CLOVER_389_0.S[8714]++;buffer.append("AdaptorRegistry[");

        __CLOVER_389_0.S[8715]++;Iterator i = registrations.entrySet().iterator();
        __CLOVER_389_0.S[8716]++;boolean showSep = false;

        __CLOVER_389_0.S[8717]++;while ((((i.hasNext()) && (++__CLOVER_389_0.CT[1466] != 0)) || (++__CLOVER_389_0.CF[1466] == 0))){
        {
            __CLOVER_389_0.S[8718]++;if ((((showSep) && (++__CLOVER_389_0.CT[1467] != 0)) || (++__CLOVER_389_0.CF[1467] == 0))){
                __CLOVER_389_0.S[8719]++;buffer.append(' ');}

            __CLOVER_389_0.S[8720]++;Map.Entry entry = (Map.Entry) i.next();

            __CLOVER_389_0.S[8721]++;Class registeredClass = (Class) entry.getKey();

            __CLOVER_389_0.S[8722]++;buffer.append(Tapestry.getClassName(registeredClass));
            __CLOVER_389_0.S[8723]++;buffer.append("=");
            __CLOVER_389_0.S[8724]++;buffer.append(entry.getValue());

            __CLOVER_389_0.S[8725]++;showSep = true;
        }}

        __CLOVER_389_0.S[8726]++;buffer.append("]");

        __CLOVER_389_0.S[8727]++;return buffer.toString();
    } finally { }}
}